#!/bin/sh
# run the /dev/mapper symlink test

# Copyright (C) 2007-2012 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

. "${abs_top_srcdir=../..}/tests/init.sh"; path_prepend_ .

. $abs_top_srcdir/tests/t-lib-helpers.sh
# Need root privileges to create a symlink under /dev/mapper.
require_root_

symlink || fail=1

Exit $fail
